from typing import Generic, TypeVar, AsyncGenerator, Callable, Self, AsyncContextManager, AsyncIterator, Any
from dataclasses import dataclass, asdict
from contextlib import asynccontextmanager
from abc import ABC, abstractmethod
from pydantic import BaseModel

from miniappi.config import settings

ConnectionT = TypeVar("ConnectionT")
SessionT = TypeVar("SessionT")

@dataclass(kw_only=True)
class UserSessionArgs:
    request_id: str

@dataclass
class Message:
    url: str
    request_id: str | None
    data: dict

@dataclass
class ServerConf:
    "Config given by Miniappi server"
    app_name: str
    app_url: str

@dataclass
class ClientConf:
    "Config given by Miniappi client"
    app_name: str | None
    version: str # Miniappi version

class AbstractUserConnection(ABC):

    @abstractmethod
    async def send(self, data: dict):
        "Send a message to a user"
        ...

    @abstractmethod
    async def listen(self) -> AsyncGenerator[Message]:
        "Listen messages from the user"
        ...

class AbstractClient(ABC):

    @abstractmethod
    async def connect_user(self, args) -> AsyncContextManager[AbstractUserConnection]:
        "Connect with a user"
        ...

    @abstractmethod
    async def listen_app(self, config: ClientConf, setup_start: Callable[[ServerConf, ...], Any]) -> AsyncIterator[Message]:
        "Connect the app and listen session starts"
        ...



import logging
from enum import StrEnum

class Loggers(StrEnum):
    user_connection = "miniappi.core.connection.user"
    app_connection = "miniappi.core.connection.app"
    connection = "miniappi.core.connection"



import asyncio
from uuid import uuid4
from typing import Dict, List, Tuple, AsyncGenerator, AsyncIterator, AsyncContextManager, Callable, Any
from collections import defaultdict
from dataclasses import dataclass, asdict
from contextlib import asynccontextmanager
from abc import ABC, abstractmethod
from .base import (
    AbstractClient, AbstractUserConnection,
    ClientConf, ServerConf,
    Message, UserSessionArgs
)

from miniappi.config import settings

@dataclass(kw_only=True)
class MockUserSessionArgs(UserSessionArgs):
    user_url: str

class MockUserConnection(AbstractUserConnection):

    def __init__(self, client: "MockClient", start_args: MockUserSessionArgs):
        self._client = client
        self._start_args = start_args

    async def send(self, data: dict):
        "Send a message to a user"
        url = self._start_args.user_url
        await self._client.response_queue[
            url
        ].put(
            data
        )

    async def listen(self) -> AsyncGenerator[Message]:
        "Listen messages from the user"
        request_id = self._start_args.request_id
        url = self._start_args.user_url
        while True:
            queue = self._client.request_queue
            this_queue = queue.get(
                url
            )
            if this_queue is not None:
                next_msg = await this_queue.get()
                msg = Message(
                    url=url,
                    request_id=request_id,
                    data=next_msg
                )
                self._client.received.append(msg)
                yield msg
            await asyncio.sleep(0)

class MockClient(AbstractClient):

    def __init__(self):
        self.session_queue: Dict[str, asyncio.Queue[Message]] = defaultdict(asyncio.Queue)
        self.request_queue: Dict[str, asyncio.Queue[Message]] = defaultdict(asyncio.Queue)
        self.response_queue: Dict[str, asyncio.Queue[Message]] = defaultdict(asyncio.Queue)

        self.sent: List[Message] = []
        self.received: List[Message] = []

        # Triggers for testing
        self.network_error_on_start = False

        # Mocks need
        self._app_name = None

    @asynccontextmanager
    async def connect_user(self, args: MockUserSessionArgs):
        "Connect with a user"
        yield MockUserConnection(self, start_args=args)

    async def listen_app(self, config: ClientConf, setup_start: Callable[[ServerConf, ...], Any]) -> AsyncIterator[Message]:
        "Connect the app and listen session starts"
        app_name = str(uuid4())
        app_url = f"{settings.url_apps}/{app_name}"
        await setup_start(
            ServerConf(
                app_name=app_name,
                app_url=app_url
            )
        )

        while True:
            queue = self.session_queue
            this_queue = queue.get(self._get_url(app_name))
            if this_queue is not None:
                next_msg = await this_queue.get()

                self.received.append(next_msg)
                yield MockUserSessionArgs(**next_msg)
            await asyncio.sleep(0)

    async def add_session(self, app_name: str, request_id: str):
        "Simulate other side of the stream"
        await self.session_queue[self._get_url(app_name)].put(
            asdict(
                MockUserSessionArgs(
                    request_id=request_id,
                    user_url=self._get_url(app_name, request_id)
                )
            )
        )

    async def _add_request(self, app_name: str, request_id: str, msg: str | dict):
        await self.request_queue[self._get_url(app_name, request_id)].put(
            msg
        )

    def _get_url(self, app_name: str, request_id: str | None = None):
        if request_id is not None:
            return f"{settings.url_start}/{app_name}/{request_id}"
        else:
            return f"{settings.url_start}/{app_name}/{request_id}"

    def _get_queue(self, app_name: str, request_id: str) -> Tuple[asyncio.Queue, asyncio.Queue]:
        url = self._get_url(app_name, request_id)
        return (
            self.request_queue.get(url),
            self.response_queue.get(url),
        )



import asyncio
import sys
import logging
from dataclasses import asdict
from contextlib import ExitStack, contextmanager
from types import TracebackType
from typing import List, Dict, Any, Awaitable, Type, Generic, TypeVar, ContextManager
from functools import partial
from collections.abc import Callable

from miniappi.core.models.callbacks import OnMessageConfig, OnOpenConfig
from miniappi.core.exceptions import UserLeftException, CloseStreamException
from . import connection as conn
from .connection import Message, ServerConf, ClientConf, UserSessionArgs
from miniappi.config import settings
from miniappi.core.models.context import Context
from miniappi.core.context import app_context as default_app_context, user_context as default_user_context
from .session import Session

from rich import print
from rich.panel import Panel

type RequestStreams = List[Callable[[dict], Awaitable[Any]]]

type CloseCallbackWithError = Callable[[type[BaseException], BaseException, TracebackType], Awaitable[Any]]
type CloseCallbackNoError = Callable[[None, None, None], Awaitable[Any]]
type CloseCallback = CloseCallbackWithError | CloseCallbackNoError


class App:
    """Miniappi application

    This class is the entrypoint
    to create apps and it handles
    the communication with Miniappi
    server.

    Args:
        app_name (str, optional):
            Name of the application. None for
            anonymous applications. Does nothing
            currently but exists for forward
            compatibility.
        user_context (Context, optional): 
            Optional extra user scoped context
            to keep track on user level
            data. This is automatically
            scoped and can be set globally.
        app_context (Context, optional):
            Optional app scoped context
            to keep track on app level data.
            This is automatically
            scoped and can be set globally.

    Examples:
        ```python
        from miniappi import App

        app = App()

        @app.on_open()
        async def new_user(session):
            print("New user joined")
            ...

        app.run()
        ```
    """

    cls_session: Type[Session] = Session

    conn_client: conn.base.AbstractClient = conn.websocket.WebsocketClient()

    def __init__(self, app_name: str | None = None,
                 user_context: Context | None = None,
                 app_context: Context | None = None):
        self.app_name = app_name

        self.callbacks_start = []
        self.callbacks_message: List[Callable[[Message], Awaitable[Any]]] = []
        self.callbacks_open: List[Callable[..., Awaitable[Any]]] = []
        self.callbacks_close: List[CloseCallback] = []
        self.callbacks_end: List[CloseCallback] = []

        self.app_context_managers: List[ContextManager] = []
        self.channel_context_managers: List[ContextManager] = []

        self.sessions: Dict[str, Session] = {}
        self.is_running = False

        self.channel_context = user_context
        self.app_context = app_context

    def _get_client_config(self):
        return ClientConf(
            app_name=self.app_name,
            version=settings.version
        )

    def get_channel_context_managers(self, session: Session):
        init_args = dict(
            session=session,
            request_id=session.start_args.request_id,
        )
        ctx = [
            *self.channel_context_managers,
            default_user_context.enter(
                init_args
            ),
        ]
        if self.channel_context:
            ctx.append(self.channel_context.enter())
        return ctx

    def get_app_context_managers(self, conf: ServerConf):
        init_args = dict(
            app=self,
            sessions=self.sessions,
            **asdict(conf)
        )
        ctx = [
            *self.app_context_managers,
            default_app_context.enter(init_args),
        ]
        if self.app_context:
            ctx.append(self.app_context.enter())
        return ctx


    def show_app_running(self, server_conf: ServerConf):
        url = server_conf.app_url
        print(
            Panel(
                "Miniappi is running.\n"
                f"[bold red]App link:[/bold red] [link={url}]{url}[/link]"
            )
        )

    async def setup_start(self, server_conf: ServerConf, echo_link: bool | None, task_group: asyncio.TaskGroup, stack: ExitStack):
        echo_link = settings.echo_url if echo_link is None else echo_link
        if echo_link:
            self.show_app_running(server_conf)
        for app_context in self.get_app_context_managers(server_conf):
            stack.enter_context(app_context)
        self.is_running = True
        self.app_name = server_conf.app_name
        await self._run_start(task_group)

    async def _listen_start(self, echo_link: bool | None, app_stack: ExitStack, task_group: asyncio.TaskGroup):
        app_config = self._get_client_config()

        setup_start = partial(self.setup_start, echo_link=echo_link, stack=app_stack, task_group=task_group)
        async for msg in self.conn_client.listen_app(app_config, setup_start=setup_start):
            yield msg

    def run(self, echo_link: bool | None = None):
        "Run app (sync)"
        asyncio.run(self.start(echo_link=echo_link))

    async def start(self, echo_link: bool | None = None):
        "Start app async"
        logger = self.get_logger("init")

        with ExitStack() as app_stack:
            try:
                async with asyncio.TaskGroup() as tg:
                    async for start_args in self._listen_start(echo_link=echo_link, app_stack=app_stack, task_group=tg):
                        tg.create_task(self.open_session(start_args))
            except ExceptionGroup as exc:
                # Exception is ExceptionGroup[ExceptionGroup]
                # Check if all expected
                await self._run_end()
                if UserLeftException._only_this(exc):
                    # Is expected
                    return
                raise
            else:
                await self._run_end()
            finally:
                self.is_running = False

    async def open_session(self, start_args: UserSessionArgs):
        logger = self.get_logger("session")
        async with self.conn_client.connect_user(start_args) as conn:
            session = self.cls_session(
                start_conn=conn,
                start_args=start_args,
                callbacks_message=self.callbacks_message,

                sessions=self.sessions
            )

            with ExitStack() as channel_stack:
                for channel_context in self.get_channel_context_managers(session):
                    channel_stack.enter_context(channel_context)
                try:
                    async with asyncio.TaskGroup() as tg:
                        session.tasks.append(tg.create_task(session.listen()))
                        for stream in self.callbacks_open:
                            args = []
                            if stream.pass_session:
                                args.append(session)
                            session.tasks.append(tg.create_task(stream(*args)))
                        logger.info(f"Session opened for client: {start_args.request_id}")
                except* UserLeftException as exc:
                    logger.info("User session closed by the user")
                    await session.close(send_stop=False)
                else:
                    await session.close()
                finally:
                    await self._run_close()

    async def _run_start(self, tg: asyncio.TaskGroup):
        for cb in self.callbacks_start:
            tg.create_task(cb())

    async def _run_end(self):
        for cb in self.callbacks_end:
            await cb(*sys.exc_info())

    async def _run_close(self):
        for cb in self.callbacks_close:
            await cb(*sys.exc_info())

    def get_logger(self, name: str):
        name = f".{name}" if name else ""
        return logging.getLogger(__name__ + name)

    @contextmanager
    def temp(self):
        """Temporarily set callbacks to the app
        
        This is useful for syncing other sessions

        Examples
        --------
        ```python
        async with app.temp() as temp:

            @temp.on_open()
            async def new_users(session):
                ...

            # Callback "new_users" will called
            # now for all new user sessions
            ...

        # Callback "new_users" won't be called
        # anymore
        ...
        ```
        """
        from miniappi.flow.app import temp_app
        with temp_app(self) as t:
            yield t

    def on_message(self):
        """Callback for user sending a message (data).

        Examples
        --------
        ```python
        @app.on_message()
        async def new_message(msg):
            ...
        ```
        """
        def wrapper(func: Callable[[dict], Awaitable[Any]]):
            self.callbacks_message.append(
                OnMessageConfig(
                    func=func,
                )
            )
            return func
        return wrapper

    def on_start(self):
        """Callback for the app starting.

        Examples
        --------
        ```python
        @app.on_start()
        async def app_start():
            ...
        ```
        """
        def wrapper(func: Callable[..., Awaitable[Any]]):
            self.callbacks_start.append(func)
            return func
        return wrapper

    def on_open(self, pass_session=False):
        """Callback for user opening an app
        session (user connected to the app)

        Examples
        --------
        ```python
        @app.on_open()
        async def new_user(session):
            ...
        ```
        """
        def wrapper(func: Callable[..., Awaitable[Any]]):
            self.callbacks_open.append(
                OnOpenConfig(
                    func=func,
                    pass_session=pass_session
                )
            )
            return func
        return wrapper

    def on_close(self):
        """Callback for user closing the
        app session (user left the app).

        Examples
        --------
        ```python
        @app.on_close()
        async def user_left(exc_type, exc, tb):
            ...
        ```
        """
        def wrapper(func: CloseCallback) -> CloseCallback:
            self.callbacks_close.append(func)
            return func
        return wrapper

    def on_end(self):
        """Callback for app shutting down.

        Examples
        --------
        ```python
        @app.on_end()
        async def app_shutdown(exc_type, exc, tb):
            ...
        ```
        """
        def wrapper(func: CloseCallback):
            self.callbacks_end.append(func)
            return func
        return wrapper



import asyncio
import logging
from typing import List, Dict, Awaitable, Any, Callable
from collections.abc import Callable
from contextlib import asynccontextmanager, AsyncExitStack
from pydantic import BaseModel
from .exceptions import UserLeftException
from .connection import AbstractUserConnection, Message
from .connection import UserSessionArgs
from .models.message_types import InputMessage, PutRoot, BaseMessage
from .models.content import BaseContent

type RequestStreams = List[Callable[[dict], Awaitable[Any]]]

class Session:

    callbacks_message: RequestStreams
    tasks: List[asyncio.Task]

    def __init__(self, start_conn: AbstractUserConnection,
                 start_args: UserSessionArgs,
                 callbacks_message: RequestStreams,
                 sessions: Dict[str, "Session"]):
        self.start_conn = start_conn
        self.start_args = start_args

        self.callbacks_message = callbacks_message

        self._sessions = sessions
        self._sessions[start_args.request_id] = self

        self.tasks = []

        self.is_running = False

    @property
    def request_id(self):
        return self.start_args.request_id

    async def send(self, data: dict | BaseModel):
        "Send to the response channel"

        logger = self.get_logger()
        logger.info("Sending data")
        body = self._format_send_message(data)

        await self.start_conn.send(body)

    def _format_send_message(self, data):
        if isinstance(data, BaseContent):
            # Considering as put message
            data = PutRoot(
                data=data
            )
        if isinstance(data, dict):
            data = InputMessage(**data)
        if not isinstance(data, BaseMessage):
            raise TypeError(f"Expected: {BaseMessage!r}, given: {type(data)!r}")
        
        return data.model_dump(exclude_none=True)

    async def _publish(self, body):
        await self.start_conn.send(body)

    async def listen(self):
        "Listen the request channel"
        logger = self.get_logger()
        try:
            logger.info("Listening channel")
            self.is_running = True
            async for message in self.start_conn.listen():
                if message is not None:
                    await self._handle_request_message(message)
                await asyncio.sleep(0)
        finally:
            self.is_running = False

    async def close(self, send_stop=True):
        "Close listening and remove session"
        logger = self.get_logger()
        logger.debug("Closing channel")
        self._sessions.pop(self.start_args.request_id)
        for task in self.tasks:
            task.cancel()
        if send_stop:
            await self._send_stop()

    async def _handle_request_message(self, msg: Message):
        if self._is_stop_message(msg):
            raise UserLeftException("Client requested to close")
        for func in self.callbacks_message:
            await func(msg)

    def get_logger(self):
        return logging.getLogger(__name__)

    def _is_stop_message(self, message: Message):
        return message.data == "OFF"

    async def _send_stop(self):
        "Send stop message"
        await self._publish("OFF")



# AUTOGENERATED FILE


from __future__ import annotations

from typing import List, Literal, Optional

from miniappi.core.models.content import BaseContent
from pydantic import ConfigDict, Field, RootModel


class Props(BaseContent):
    model_config = ConfigDict(
        extra='allow',
    )
    id: str = Field(..., description='ID of the component')
    options: List[str]
    placeholder: Optional[str] = None
    submitText: Optional[str] = None
    contentType: Literal['v0/widgets/DropDown.vue'] = 'v0/widgets/DropDown.vue'


class Root(RootModel[Props]):
    root: Props



# AUTOGENERATED FILE


from __future__ import annotations

from typing import List, Literal, Optional, Union

from miniappi.core.models.content import BaseContent
from pydantic import ConfigDict, Field, RootModel


class DatePicker(BaseContent):
    model_config = ConfigDict(
        extra='allow',
    )
    groupId: Optional[str] = None
    label: str
    name: str
    required: Optional[bool] = None
    type: Literal['DatePicker'] = 'DatePicker'


class Email(BaseContent):
    model_config = ConfigDict(
        extra='allow',
    )
    groupId: Optional[str] = None
    label: str
    name: str
    required: Optional[bool] = None
    type: Literal['email'] = 'email'


class InputNumber(BaseContent):
    model_config = ConfigDict(
        extra='allow',
    )
    groupId: Optional[str] = None
    label: str
    name: str
    required: Optional[bool] = None
    type: Literal['InputNumber'] = 'InputNumber'


class InputText(BaseContent):
    model_config = ConfigDict(
        extra='allow',
    )
    groupId: Optional[str] = None
    label: str
    name: str
    required: Optional[bool] = None
    type: Literal['InputText'] = 'InputText'


class Password(BaseContent):
    model_config = ConfigDict(
        extra='allow',
    )
    groupId: Optional[str] = None
    label: str
    name: str
    required: Optional[bool] = None
    type: Literal['password'] = 'password'


class Rating(BaseContent):
    model_config = ConfigDict(
        extra='allow',
    )
    groupId: Optional[str] = None
    label: str
    name: str
    required: Optional[bool] = None
    stars: Optional[float] = None
    type: Literal['Rating'] = 'Rating'


class Props(BaseContent):
    model_config = ConfigDict(
        extra='allow',
    )
    fields: List[Union[Email, Password, InputText, InputNumber, Rating, DatePicker]]
    id: str = Field(..., description='ID of the component')
    legend: Optional[str] = None
    contentType: Literal['v0/forms/DynamicForm.vue'] = 'v0/forms/DynamicForm.vue'


class Root(RootModel[Props]):
    root: Props



from .button import Props as Button
from .drop_down import Props as DropDown
from .editor import Props as Editor
from .input_number import Props as InputNumber
from .input_text import Props as InputText
from .knob import Props as Knob



import asyncio
from typing import Generic, TypeVar
from collections import UserList
from miniappi.core import user_context, app_context, Session
from miniappi.core.models.message_types import PushRight, PutRef

T = TypeVar("T")

class Feed(UserList, Generic[T]):
    """Content Feed"""

    def __init__(self, initlist=None, id=None):
        super().__init__(initlist)
        self.id = id

    async def _push_session(self, elem, session: Session):
        await session.send(
            PushRight(
                id=self.id,
                data=elem
            )
        )

    async def show(self, session: Session | None = None):
        "Send all items in the feed to the session"
        msg = PutRef(
            id=self.id,
            data=self.data
        )
        if session is not None:
            return await session.send(msg)
        try:
            session = user_context.session
            await session.send(msg)
        except LookupError:
            # Called outside of channel
            # --> set as root to all channels
            for session in app_context.sessions.values():
                await session.send(msg)

    async def append_all(self, element: T):
        "Append to the feed and show it to all"
        self.data.append(element)
        for session in app_context.sessions.values():
            await self._push_session(element, session)

    async def append(self, element: T):
        """Append to the feed and show it to the user
        (if user context) or all (if no user context)"""
        self.data.append(element)
        try:
            await self._push_session(element, user_context.session)
        except LookupError:
            for session in app_context.sessions.values():
                await self._push_session(element, session)

    def as_reference(self):
        return {
            "reference": self.id,
            "data": self.data
        }



import asyncio
from typing import Generic, TypeVar, Literal
from collections import UserList
from miniappi.core import BaseContent, user_context, app_context
from miniappi.core.connection import Message
from miniappi.flow import temp_app, in_channel_scope, in_app_scope

async def wait_for_input(content: BaseContent, show: bool = False, any_child=True, wait_for: Literal["any", "all"] | None = None):
    """Wait for input to the content
    
    If not in channel context, wait for all input"""
    def set_if_ready():
        if wait_for == "all" or (wait_for is None and not in_channel):
            for session in app_context.sessions.values():
                if session.request_id not in outputs:
                    # Not all ready, don't set the event
                    return
        event.set()

    def get_child_ids():
        return [c.id for c in content.iter_content()]

    in_channel = in_channel_scope()

    only_caller = in_channel and wait_for is None
    caller_request_id = user_context.request_id if in_channel else None

    ids = content.id if not any_child else get_child_ids() + [content.id]
    event = asyncio.Event()
    outputs = {}
    with temp_app() as app:
        @app.on_message()
        async def get_message(msg: Message):
            data: dict = msg.data
            if data.get("id") in ids:
                if only_caller and msg.request_id != caller_request_id:
                    return
                # We don't need the content ID with
                # the msg so we just return data
                outputs[msg.request_id] = data
                set_if_ready()
        if show:
            await content.show()
        await event.wait()

    if only_caller:
        return outputs[caller_request_id]
    return outputs




class StreamException(Exception):
    ...

    @classmethod
    def _only_this(cls, e_group: ExceptionGroup):
        for exc in e_group.exceptions:
            if not isinstance(exc, cls):
                return False
            if isinstance(exc, ExceptionGroup):
                if not cls._only_this(exc):
                    return False
        return True

class UserLeftException(StreamException):
    ...

class CloseStreamException(StreamException):
    ...



import time
from datetime import timedelta
import asyncio
from typing import Callable, List, AsyncIterable
from contextlib import asynccontextmanager
from uuid import uuid4

from miniappi import app_context
from miniappi.core import App
from miniappi.core.connection import Message
from miniappi.core.connection.mock import MockClient, MockUserSessionArgs

async def _wait_for_condition(cond: Callable, sleep=0, timeout: int=5):
    start = time.time()
    while not cond():
        if (time.time() - start) > timeout:
            raise TimeoutError(cond)
        await asyncio.sleep(sleep)

class StreamHandler:

    conn_client: MockClient
    received: List[Message]
    sent: List[Message]

    def __init__(self, stream: App, request_id: str):
        self.stream = stream
        self.request_id = request_id

        self.received = []
        self.sent = []

        self._n_sent = 0

        self.conn_client = self.stream.conn_client
        if not isinstance(self.stream.conn_client, MockClient):
            raise TypeError("Testing handler only works for MockClient")

    @property
    def start_args(self):
        return MockUserSessionArgs(
            request_id=self.request_id,
            channel=self.channel_name,
        )

    async def start_communication(self, timeout):
        await self.conn_client.add_session(
            app_name=self.stream.app_name,
            request_id=self.request_id
        )
        async with asyncio.timeout(timeout):
            while not self.is_running():
                await asyncio.sleep(0)

    def is_running(self):
        session = self.stream.sessions.get(self.request_id)
        if not session:
            return False
        return session.is_running

    def get_url(self):
        return self.conn_client._get_url(
            self.stream.app_name,
            self.request_id
        )

    async def wait_for_sent(self):
        req, resp = self.conn_client._get_queue(
            self.stream.app_name,
            self.request_id
        )
        if resp is None:
            return
        while not resp.empty():
            message_data = await resp.get()
            message = Message(
                url=self.stream.app_name,
                request_id=self.request_id,
                data=message_data
            )
            self.sent.append(message)
            await asyncio.sleep(0)

    async def wait_for_received(self):
        while self._n_sent != len(self.received):
            await asyncio.sleep(0)

    async def wait_for_messages(self):
        await self.wait_for_received()
        await self.wait_for_sent()

    async def get_next_sent(self):
        "Get next sent message"
        while True:
            queue = self.conn_client.response_queue.get(
                (self.stream.app_name, self.request_id)
            )
            if queue is not None:
                while not queue.empty():
                    message_data = await queue.get()
                    message = Message(
                        url=self.channel_name,
                        request_id=self.request_id,
                        data=message_data
                    )
                    self.sent.append(message)
                    return message
            await asyncio.sleep(0)

    async def send_message(self, msg: dict | str):
        "Send message to the streamer"
        await self.conn_client._add_request(
            app_name=self.stream.app_name,
            request_id=self.request_id,
            msg=msg
        )
        self._n_sent += 1

@asynccontextmanager
async def listen(stream: App, request_id: str = None, start_args: dict = None, wait_close=False):
    "Communicate with a stream"
    start_args = start_args or {}
    request_id = request_id or str(uuid4())
    timeout = 10

    handler = StreamHandler(
        stream=stream,
        request_id=request_id,
    )

    @stream.on_message()
    async def record_message(message: Message):
        if message.url == handler.get_url():
            handler.received.append(message)

    async with asyncio.timeout(timeout):
        while True:
            if stream.is_running:
                break
            await asyncio.sleep(0)

    await handler.start_communication(
        timeout=timeout
    )

    async with asyncio.timeout(timeout):
        yield handler

    async with asyncio.timeout(timeout):
        await handler.wait_for_messages()

    if wait_close:
        async with asyncio.timeout(timeout):
            while stream.is_running:
                await asyncio.sleep(0)



from typing import Dict
from miniappi.core.context import CurrentContent
from miniappi.core.models.content import BaseContent
from miniappi.core.models.message_types import InputMessage, PutRoot

def handle_message(curr_content: CurrentContent, msg: InputMessage):

    if msg.type == "root":
        curr_content.root = msg.data
    elif msg.type == "ref":
        reference = curr_content.references[msg.id]

        if msg.method == "put":
            curr_content.references[msg.id] = msg.data
        elif msg.method == "delete":
            curr_content.references.pop(msg.id, None)
        elif msg.method == "push":
            push_reference(curr_content.references, msg)
        elif msg.method == "pop":
            pop_reference(curr_content.references, msg)
        else:
            raise ValueError(f"unrecognized method: {msg.method}")

def push_reference(references: Dict[str, any], msg: InputMessage):
    reference = references[msg.id]
    if isinstance(reference, list):
        if msg.key is None:
            reference.append(msg.data)
        else:
            reference.insert(msg.key, msg.data)
    elif isinstance(reference, dict):
        if msg.key is None:
            raise KeyError("Key undefined")
        else:
            reference[msg.key] = msg.data

def pop_reference(references: Dict[str, any], msg: InputMessage):
    reference = references[msg.key]
    if isinstance(msg.key, int):
        reference: list
        reference.pop(msg.key)
    elif isinstance(reference, dict):
        reference.pop(msg.key)



# AUTOGENERATED FILE


from __future__ import annotations

from typing import Literal, Optional

from miniappi.core.models.content import BaseContent
from pydantic import ConfigDict, Field, RootModel


class Props(BaseContent):
    model_config = ConfigDict(
        extra='allow',
    )
    defaultValue: Optional[str] = Field(
        None,
        description='The default value for the input when not controlled by `modelValue`.',
    )
    id: str = Field(..., description='ID of the component')
    inputmode: Optional[
        Literal['none', 'text', 'tel', 'url', 'email', 'numeric', 'decimal', 'search']
    ] = Field(
        None,
        description='Hints at the type of data that might be entered by the user while editing the element or its contents',
    )
    placeholder: Optional[str] = None
    submitText: Optional[str] = None
    contentType: Literal['v0/widgets/InputText.vue'] = 'v0/widgets/InputText.vue'


class Root(RootModel[Props]):
    root: Props



# AUTOGENERATED FILE
# generated by datamodel-codegen:


from __future__ import annotations

from typing import Any, Literal, Optional, Union

from miniappi.core.models.content import BaseMessage
from pydantic import ConfigDict, RootModel


class Root(RootModel[Any]):
    root: Any


class DelRef(BaseMessage):
    model_config = ConfigDict(
        extra='allow',
    )
    id: str
    method: Literal['delete'] = 'delete'
    type: Literal['ref'] = 'ref'


class InputMessage(BaseMessage):
    """
    Message interface
    """

    model_config = ConfigDict(
        extra='allow',
    )
    data: Optional[Any] = None
    id: Optional[str] = None
    key: Optional[Union[str, float]] = None
    method: Optional[Literal['put', 'delete', 'push', 'pop']] = None
    type: Literal['root', 'ref']


class PopLeft(BaseMessage):
    model_config = ConfigDict(
        extra='allow',
    )
    id: str
    key: Literal[-1] = -1
    method: Literal['pop'] = 'pop'
    type: Literal['ref'] = 'ref'


class PopRef(BaseMessage):
    model_config = ConfigDict(
        extra='allow',
    )
    id: str
    key: Union[str, float]
    method: Literal['pop'] = 'pop'
    type: Literal['ref'] = 'ref'


class PopRight(BaseMessage):
    model_config = ConfigDict(
        extra='allow',
    )
    id: str
    key: Literal[0] = 0
    method: Literal['pop'] = 'pop'
    type: Literal['ref'] = 'ref'


class PushLeft(BaseMessage):
    model_config = ConfigDict(
        extra='allow',
    )
    data: Any
    id: str
    key: Literal[0] = 0
    method: Literal['push'] = 'push'
    type: Literal['ref'] = 'ref'


class PushRef(BaseMessage):
    model_config = ConfigDict(
        extra='allow',
    )
    data: Any
    id: str
    key: Optional[Union[str, float]] = None
    method: Literal['push'] = 'push'
    type: Literal['ref'] = 'ref'


class PushRight(BaseMessage):
    model_config = ConfigDict(
        extra='allow',
    )
    data: Any
    id: str
    key: Optional[Any] = None
    method: Literal['push'] = 'push'
    type: Literal['ref'] = 'ref'


class PutRef(BaseMessage):
    model_config = ConfigDict(
        extra='allow',
    )
    data: Any
    id: str
    method: Literal['put'] = 'put'
    type: Literal['ref'] = 'ref'


class PutRoot(BaseMessage):
    model_config = ConfigDict(
        extra='allow',
    )
    data: Any
    method: Literal['put'] = 'put'
    type: Literal['root'] = 'root'


class ShowRef(BaseMessage):
    model_config = ConfigDict(
        extra='allow',
    )
    id: str
    method: Literal['show'] = 'show'
    type: Literal['ref'] = 'ref'


class ShowRoot(BaseMessage):
    model_config = ConfigDict(
        extra='allow',
    )
    id: str
    method: Literal['show'] = 'show'
    type: Literal['root'] = 'root'



from typing import Union, TYPE_CHECKING, Generator
from uuid import uuid4
from contextlib import contextmanager
from pydantic import BaseModel
from ..context import user_context, app_context

if TYPE_CHECKING:
    from ..app.stream import AppSession

class BaseContent(BaseModel):
    id: str

    def __init__(self, id: str | None = None, *args, **kwargs):
        if id is None:
            id = str(uuid4())
        super().__init__(*args, id=id, **kwargs)

    async def show(self, session: Union["AppSession", None] = None):
        if session is not None:
            return await session.send(self)
        try:
            session = user_context.session
            await session.send(self)
        except LookupError:
            # Called outside of channel
            # --> set as root to all channels
            for session in app_context.sessions.values():
                await session.send(self)

    async def wait_input(self, show=True, *args, **kwargs):
        from miniappi.flow.interact import wait_for_input
        output = await wait_for_input(self, show=show, *args, **kwargs)
        return output

    def iter_content(self) -> Generator["BaseContent"]:

        def _iter_content(value):
            if isinstance(value, BaseContent):
                yield from value.iter_content()
            elif isinstance(value, (list, tuple, set)):
                for subval in value:
                    yield from _iter_content(subval)
            elif isinstance(value, dict):
                for subval in value.values():
                    yield from _iter_content(subval)

        yield self
        fields = type(self).model_fields
        for name, field in fields.items():
            value = getattr(self, name)
            yield from _iter_content(value)

class BaseMessage(BaseModel):
    ...



# AUTOGENERATED FILE


from __future__ import annotations

from typing import Literal

from miniappi.core.models.content import BaseContent
from pydantic import ConfigDict, Field, RootModel


class Props(BaseContent):
    model_config = ConfigDict(
        extra='allow',
    )
    code: str
    id: str = Field(..., description='ID of the component')
    language: Literal[
        'bash',
        'javascript',
        'python',
        'java',
        'c',
        'cpp',
        'rust',
        'r',
        'sql',
        'scala',
        'html',
        'json',
        'yaml',
        'markdown',
    ]
    contentType: Literal['v0/CodeBlock.vue'] = 'v0/CodeBlock.vue'


class Root(RootModel[Props]):
    root: Props



# AUTOGENERATED FILE


from __future__ import annotations

from typing import Literal

from miniappi.core.models.content import BaseContent
from pydantic import ConfigDict, Field, RootModel


class Props(BaseContent):
    """
    Simple large header
    """

    model_config = ConfigDict(
        extra='allow',
    )
    id: str = Field(..., description='ID of the component')
    text: str
    contentType: Literal['v0/Title.vue'] = 'v0/Title.vue'


class Root(RootModel[Props]):
    root: Props



from pathlib import Path
from importlib.metadata import version, PackageNotFoundError
from pydantic_settings import BaseSettings, SettingsConfigDict

class Settings(BaseSettings):
    "Miniappi configuration"
    model_config = SettingsConfigDict(
        env_prefix='miniappi_'
    )

    url_start: str = "https://miniappi.com/api/v1/streams/apps/start"
    url_recover: str = "https://miniappi.com/api/v1/streams/apps/recover"
    url_apps: str = "https://miniappi.com/apps"

    echo_url: bool | None = True

    keepalive_ping_interval: float | None = 20.0
    keepalive_ping_timeout: float | None = 20.0
    timeout: float | None = None

    @property
    def version(self):
        "Version of Miniappi"
        try:
            return version("miniappi")
        except PackageNotFoundError:
            # Probably run in dev
            return "0.0.0"

settings = Settings()



from typing import Any, Awaitable
from collections.abc import Callable
from pydantic import BaseModel
from miniappi.core.connection import Message

class StartArgs(BaseModel):
    request_id: str
    request_channel: str | None
    response_channel: str | None
    channel: str | None = None

class OnMessageConfig:

    def __init__(self, func: Callable[[Message], Awaitable[Any]]):
        self.func = func

    async def __call__(self, msg: Message):
        return await self.func(msg)

class OnOpenConfig:

    def __init__(self, func: Callable[[Message], Awaitable[Any]], pass_session):
        self.func = func
        self.pass_session = pass_session

    async def __call__(self, *args, **kwargs):
        return await self.func(*args, **kwargs)



import inspect
import dataclasses
from uuid import uuid4
from typing import Any, Optional, Generic, TypeVar, Dict
from contextvars import copy_context
from collections import UserDict
from contextlib import contextmanager
from contextvars import ContextVar, Token

ContextDataT = TypeVar("ContextDataT")

class Context(Generic[ContextDataT]):

    def __init__(self, name: str):
        self.store: ContextVar[ContextDataT] = ContextVar(
            name
        )

    @property
    def data(self) -> ContextDataT:
        return self.store.get()

    def exists(self) -> bool:
        return self.store in copy_context()

    @contextmanager
    def enter(self, data: ContextDataT):
        "Enter context"
        if self.exists():
            raise LookupError("Context already initiated")

        token: Token = self.store.set(data)
        yield self.store
        self.store.reset(token)

    def __repr__(self) -> str:
        try:
            return f"<{__name__}.{self.__class__.__name__} {self.data}>"
        except LookupError:
            return f"<{__name__}.{self.__class__.__name__}>"

    def __str__(self) -> str:
        try:
            return str(self.data)
        except LookupError:
            return str({})

class ContextDict(Context[Dict[Any, Any]], UserDict):

    def enter(self, data: Optional[Dict] = None):
        if data is None:
            data = {}
        return super().enter(data.copy())

MISSING = dataclasses.MISSING

class ContextModel(Context[Dict[Any, Any]]):
    __ignore_extra__: bool = False

    def __init__(self, name_: str | None = None, **kwargs):
        if name_ is None:
            name_ = str(uuid4())
        self.store: ContextVar[ContextDataT] = ContextVar(
            name_
        )
        self._set_attrs(**kwargs)

    def _set_attrs(self, **kwargs):
        attrs = inspect.get_annotations(type(self))
        fields: Dict[str, dataclasses.Field] = {}
        for name, type_ in attrs.items():
            try:
                default = super().__getattribute__(name)
            except AttributeError:
                default = MISSING
            if isinstance(default, dataclasses.Field):
                fields[name] = default
            else:
                field = dataclasses.field()
                field.name = name
                field.default = default
                field.type = type_
                fields[name] = field
            if name in kwargs:
                field.default = kwargs[name]
        self.__fields__ = fields

    def enter(self, data: Optional[Dict] = None):
        defaults = {
            name: field.default if field.default is not MISSING else field.default_factory()
            for name, field in self.__fields__.items()
            if (
                field.default is not MISSING
                or field.default_factory is not MISSING
            )
        }
        if data is None:
            data = {}
        elif not self.__ignore_extra__:
            extra_fields = [
                field
                for field in data
                if field not in self.__fields__
            ]
            if extra_fields:
                raise TypeError(f"'{type(self)}' has no attribute(s): {', '.join(extra_fields)}")
        data = {**defaults, **data}

        missing_fields = [
            field_name
            for field_name in self.__fields__
            if field_name not in data
        ]
        if missing_fields:
            raise TypeError(f'Missing required argument(s): {", ".join(missing_fields)}')
        return super().enter(data)

    def __getattribute__(self, key: str):
        try:
            fields = super().__getattribute__("__fields__")
        except AttributeError:
            return super().__getattribute__(key)

        if key in fields:
            value = self.data[key]
            return value
        return super().__getattribute__(key)

    def __setattr__(self, name: str, value):
        if name in ("store",) or name.startswith("_"):
            return super().__setattr__(name, value)
        if name not in self.__fields__:
            self._raise_attr_error(name)
        self.data[name] = value

    def _raise_attr_error(self, name: str):
        raise AttributeError(f"'{type(self)!s}' object has no attribute '{name}'")



from contextlib import asynccontextmanager
import json
import logging
import asyncio
from typing import Callable, Any
from dataclasses import dataclass, asdict
from pydantic import BaseModel
from httpx_ws import aconnect_ws, AsyncWebSocketSession, WebSocketNetworkError, WebSocketDisconnect
from httpx import AsyncClient

from miniappi.core.logging import Loggers

from miniappi.core.exceptions import UserLeftException
from .base import (
    AbstractUserConnection,
    AbstractClient,
    ServerConf,
    ClientConf,
    UserSessionArgs,
    Message
)
from miniappi.config import settings

LOGGER = logging.getLogger(Loggers.connection.value)
LOGGER_USER = logging.getLogger(Loggers.user_connection.value)
LOGGER_APP = logging.getLogger(Loggers.app_connection.value)

@dataclass(kw_only=True)
class WebsocketUserSessionArgs(UserSessionArgs):
    user_url: str

@dataclass
class RecoveryConf:
    recovery_key: str

async def _listen_messages(ws: AsyncWebSocketSession):
    while True:
        data = await ws.receive_text()
        if data.lower() == "off":
            # Users disconnected
            LOGGER.debug("Received a close message")
            raise UserLeftException("Server closed the session")
        elif data.lower() == "ping":
            LOGGER.debug("Received a ping message")
            continue
        message = json.loads(data)
        yield message
        await asyncio.sleep(0)

class WebsocketUserConnection(AbstractUserConnection):

    def __init__(self, ws: AsyncWebSocketSession, start_args: WebsocketUserSessionArgs):
        self.ws = ws
        self.start_args = start_args

    async def send(self, data: dict):
        "Send a message to a user"
        await self.ws.send_json(data)
        LOGGER_USER.info("Message sent")

    async def listen(self):
        "Listen messages from the user"
        try:
            async for msg in _listen_messages(self.ws):
                LOGGER_USER.info("Message received")
                yield Message(
                    url=self.start_args.user_url,
                    request_id=self.start_args.request_id,
                    data=msg
                )
        except WebSocketDisconnect as exc:
            if exc.code in (1000, 1001):
                if exc.code == 1001:
                    LOGGER_USER.info("User left")
                else:
                    LOGGER_USER.info("User session closed by server")
                raise UserLeftException(exc.reason)
            LOGGER_USER.exception("Server disconnected")
            raise

class WebsocketClient(AbstractClient):

    def __init__(self, client: AsyncClient | None = None):
        self.client = client or AsyncClient(timeout=settings.timeout)

    @asynccontextmanager
    async def connect_user(self, start_args: WebsocketUserSessionArgs):
        async with aconnect_ws(
            start_args.user_url,
            client=self.client,
            keepalive_ping_interval_seconds=settings.keepalive_ping_interval,
            keepalive_ping_timeout_seconds=settings.keepalive_ping_timeout
        ) as ws:
            yield WebsocketUserConnection(ws, start_args)
            ...
        ...

    async def listen_app(self, conf: ClientConf, setup_start: Callable[[ServerConf, ...], Any]):
        url = (
            settings.url_start
            if conf.app_name is None
            else f"{settings.url_start}/{conf.app_name}"
        )
        n_fails = 0
        recovery_key = None
        is_reconnect = False
        while True:
            try:
                if is_reconnect:
                    LOGGER_APP.info("Reconnecting to app...")
                else:
                    LOGGER_APP.info("Connecting to app...")
                async with aconnect_ws(
                    url,
                    client=self.client,
                    keepalive_ping_interval_seconds=settings.keepalive_ping_interval,
                    keepalive_ping_timeout_seconds=settings.keepalive_ping_timeout
                ) as ws:
                    n_fails = 0
                    if not is_reconnect:
                        # Initialize the app
                        await ws.send_json(asdict(conf))
                        server_conf = ServerConf(**await ws.receive_json())
                        await setup_start(
                            server_conf
                        )
                    recovery_conf = RecoveryConf(**await ws.receive_json())
                    recovery_key = recovery_conf.recovery_key
                    LOGGER_APP.info("App connected")
                    async for msg in _listen_messages(ws):
                        LOGGER_APP.info("User joined")
                        yield WebsocketUserSessionArgs(**msg)
            except WebSocketNetworkError as exc:
                # Reconnecting...
                if recovery_key:
                    is_reconnect = True
                    url = f"{settings.url_recover}/{recovery_key}"
                else:
                    # Didn't yet even initialize
                    is_reconnect = False

                # Wait for 0, 1, 8, 27, 64 ...
                # after each fail in row

                reconnect_delay = (n_fails + 1) ** 3
                n_fails += 1
                if reconnect_delay > (60 * 60):
                    # Too many fails, the server timeouts
                    # before the wait
                    LOGGER_APP.exception(
                        "Network interupted. Could not reconnect"
                    )
                    raise
                
                LOGGER_APP.warning(
                    f"Network interupted. Waiting {reconnect_delay} "
                    "sec before reconnecting..."
                )
                await asyncio.sleep(reconnect_delay)
            except WebSocketDisconnect as exc:
                if exc.code in (1001, 1000):
                    # Normal closure
                    LOGGER_APP.info("Server closed")
                    raise UserLeftException(exc.reason)
                LOGGER_APP.exception("Server disconnected")
                raise



from copy import copy
from typing import TYPE_CHECKING, Any, Dict, Union
from dataclasses import dataclass, field
from miniappi.core.models.context import ContextModel

if TYPE_CHECKING:
    from .models.content import BaseContent
    from . import App, Session


@dataclass
class CurrentContent:
    root: Union["BaseContent", None] = None
    references: Dict[str, "BaseContent"] = field(default_factory=lambda: {})

class UserContext(ContextModel):

    session: "Session" # Current stream session
    request_id: str
    extra: dict = field(default_factory=lambda: {})

    def copy(self):
        return copy(self)

class AppContext(ContextModel):

    app: "App"
    app_name: str
    app_url: str
    sessions: Dict[str, "Session"]
    extra: dict = field(default_factory=lambda: {})

    def copy(self):
        return copy(self)

user_context = UserContext()
app_context = AppContext()



# AUTOGENERATED FILE


from __future__ import annotations

from typing import Literal, Optional

from miniappi.core.models.content import BaseContent
from pydantic import ConfigDict, Field, RootModel


class Props(BaseContent):
    model_config = ConfigDict(
        extra='allow',
    )
    defaultValue: Optional[float] = Field(
        None,
        description='The default value for the input when not controlled by `modelValue`.',
    )
    id: str = Field(..., description='ID of the component')
    max: Optional[float] = Field(None, description='Maximum boundary value.')
    maxFractionDigits: Optional[float] = Field(
        None,
        description="The maximum number of fraction digits to use. Possible values are from 0 to 20; the default for plain number formatting is the larger of minimumFractionDigits and 3; the default for currency formatting is the larger of minimumFractionDigits and the number of minor unit digits provided by the [ISO 4217 currency code](https://www.six-group.com/en/products-services/financial-information/data-standards.html#scrollTo=maintenance-agency) list (2 if the list doesn't provide that information).",
    )
    min: Optional[float] = Field(None, description='Minimum boundary value.')
    minFractionDigits: Optional[float] = Field(
        None,
        description="The minimum number of fraction digits to use. Possible values are from 0 to 20; the default for plain number and percent formatting is 0; the default for currency formatting is the number of minor unit digits provided by the [ISO 4217 currency code](https://www.six-group.com/en/products-services/financial-information/data-standards.html#scrollTo=maintenance-agency) list (2 if the list doesn't provide that information).",
    )
    prefix: Optional[str] = Field(None, description='Text to display before the value.')
    showButtons: Optional[bool] = Field(None, description='Displays spinner buttons.')
    step: Optional[float] = Field(
        None, description='Step factor to increment/decrement the value.'
    )
    submitText: Optional[str] = None
    suffix: Optional[str] = Field(None, description='Text to display after the value.')
    contentType: Literal['v0/widgets/InputNumber.vue'] = 'v0/widgets/InputNumber.vue'


class Root(RootModel[Props]):
    root: Props



# AUTOGENERATED FILE


from __future__ import annotations

from typing import List, Literal, Optional, Union

from miniappi.core.models.content import BaseContent
from pydantic import ConfigDict, Field, RootModel


class ReferenceAny(BaseContent):
    """
    Reference content/attribute References are useful for fast changing data to avoid sending the whole UI structure every time.
    """

    model_config = ConfigDict(
        extra='allow',
    )
    data: Optional[List] = Field(
        None, description='If given, the reference is initialized with given type'
    )
    reference: str = Field(..., description='ID of the reference')


Reference3Cany5B5D3E = ReferenceAny


class Props(BaseContent):
    model_config = ConfigDict(
        extra='allow',
    )
    cols: Optional[float] = None
    contents: Union[List, Reference3Cany5B5D3E]
    id: str = Field(..., description='ID of the component')
    rows: Optional[float] = None
    contentType: Literal['v0/layouts/Grid.vue'] = 'v0/layouts/Grid.vue'


class Root(RootModel[Props]):
    root: Props



# AUTOGENERATED FILE


from __future__ import annotations

from typing import Any, Literal, Optional

from miniappi.core.models.content import BaseContent
from pydantic import ConfigDict, Field, RootModel


class Props(BaseContent):
    model_config = ConfigDict(
        extra='allow',
    )
    body: Optional[Any] = None
    footer: Optional[Any] = None
    header: Optional[Any] = None
    id: str = Field(..., description='ID of the component')
    subtitle: Optional[str] = None
    title: Optional[str] = None
    contentType: Literal['v0/cards/Card.vue'] = 'v0/cards/Card.vue'


class Root(RootModel[Props]):
    root: Props



from contextlib import contextmanager
from miniappi.core import app_context

class TempApp:

    def __init__(self, app=None):
        self._app = app_context.app if app is None else app
        self._callbacks = {
            self._app.on_open: [],
            self._app.on_message: [],
            self._app.on_close: [],
            self._app.on_start: [],
            self._app.on_end: [],
        }

    def _get_app_callbacks(self, app_meth):
        return {
            self._app.on_open: self._app.callbacks_open,
            self._app.on_message: self._app.callbacks_message,
            self._app.on_close: self._app.callbacks_close,
            self._app.on_start: self._app.callbacks_start,
            self._app.on_end: self._app.callbacks_end,
        }[app_meth]

    def _on_callback(self, *args, _app_meth, **kwargs):
        on_init = _app_meth()
        def wrapper(func):
            out = on_init(func)
            # The app might create some wrapper
            # for the func (ie. see on_message)
            # so we get the last callback
            self._callbacks[_app_meth].append(
                self._get_app_callbacks(_app_meth)[-1]
            )
            return out
        return wrapper

    def on_start(self, *args, **kwargs):
        return self._on_callback(
            *args,
            **kwargs,
            _app_meth=self._app.on_start
        )

    def on_open(self, *args, **kwargs):
        return self._on_callback(
            *args,
            **kwargs,
            _app_meth=self._app.on_open
        )

    def on_message(self, *args, **kwargs):
        return self._on_callback(
            *args,
            **kwargs,
            _app_meth=self._app.on_message
        )

    def on_close(self, *args, **kwargs):
        return self._on_callback(
            *args,
            **kwargs,
            _app_meth=self._app.on_close
        )

    def on_end(self, *args, **kwargs):
        return self._on_callback(
            *args,
            **kwargs,
            _app_meth=self._app.on_end
        )

    def remove(self):
        "remove all set callbacks"
        for meth, setcbs in self._callbacks.items():
            for cb in setcbs:
                app_cbs = self._get_app_callbacks(meth)
                app_cbs.remove(cb)

@contextmanager
def temp_app(app=None):
    "Context manager for setting app configs temporarily (ie. callbacks)"
    handler = TempApp(app)
    yield handler
    handler.remove()



from miniappi import app_context, user_context

def in_channel_scope() -> bool:
    """Return true if in channel scope
    (called from a function of channel open)"""
    try:
        user_context.session
    except LookupError:
        return False
    else:
        return True

def in_app_scope() -> bool:
    """Return true if in app scope
    (called from a function of app running)"""
    try:
        app_context.app
    except LookupError:
        return False
    else:
        return True
